# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum

from .reader import AllowedOptions, AbstractConfigReader, IniReader, JsonReader
from .factory import ReaderFactory
from ..utils import EnumUtil, ExtensionNameList, get_from_allowed, PathTypes
from ..validator.is_validators import is_string, is_enum_subclass


class ReaderType(Enum):
    INI = IniReader
    JSON = JsonReader


class ReaderBuilder(object):
    _classes = EnumUtil.create_dict_by_enum(ReaderType)

    def __init__(self,
                 filename: PathTypes,
                 options: AllowedOptions,
                 extension_names: ExtensionNameList = None):
        self._filename = filename
        self._options = options
        self._extension_names = [] if not extension_names else extension_names

        self._reader_class = ReaderType.INI
        self._kwargs = {}

    def build(self) -> AbstractConfigReader:
        return ReaderFactory.create(
            self._get_allowed_class(self._reader_class),
            self._filename,
            self._options,
            self._extension_names,
            **self._kwargs
        )

    def set_reader(self, reader_class: ReaderType):
        if not self._is_allowed_class(reader_class):
            return self
        self._set_allowed_class(reader_class)
        return self

    def set_options(self, options: AllowedOptions):
        if not is_enum_subclass(options):
            return self
        self._options = options
        return self

    def set_kwargs(self, **kwargs):
        if not kwargs:
            return self
        self._kwargs.update(kwargs)
        return self

    def add_extension_name(self, extension_name: str):
        if not is_string(extension_name):
            return self
        self._extension_names.append(extension_name)
        return self

    def _get_allowed_class(self, reader_class: ReaderType) -> AbstractConfigReader:
        return get_from_allowed(reader_class, ReaderBuilder._classes)

    def _set_allowed_class(self, reader_class: ReaderType) -> None:
        self._reader_class = reader_class

    def _is_allowed_class(self, reader_class: ReaderType):
        return reader_class in ReaderBuilder._classes



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractCallbackManager
from ..validator import check_function, check_async_function


class CallbackManager(AbstractCallbackManager):

    def check(self, value):
        return check_function(value)

    def notify(self, *args, **kwargs):
        for cb in self:
            try:
                cb(*args, **kwargs)
            except Exception as e:
                raise e


class AsyncCallbackManager(AbstractCallbackManager):

    def check(self, value):
        return check_async_function(value)

    async def async_notify(self, *args, **kwargs):
        async for cb in self:
            try:
                await cb(*args, **kwargs)
            except Exception as e:
                raise e

    async def __aiter__(self):
        if not self.empty():
            for _, cb in enumerate(self.callbacks):
                yield cb



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum
from typing import Any

from ..validator.is_validators import is_enum, is_enum_subclass


class EnumUtil(object):

    @classmethod
    def to_value(cls, enums: Enum) -> Any:
        if not is_enum(enums):
            raise ValueError(f"enum need Enum instance, but got {type(enums).__name__}")
        return enums.value

    @classmethod
    def create_dict_by_enum(cls, enum_klass: Enum) -> dict:
        if not is_enum_subclass(enum_klass):
            raise ValueError(f"enum klass not Enum subclass, got {type(enum_klass).__name__}")
        to_value = cls.to_value
        return {item: to_value(item) for _, item in enumerate(enum_klass)}



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from typing import Any

from ..validator.is_validators import is_dict


def get_from_allowed(key: Any, allowed_props: dict) -> Any:
    allowed = {} if not is_dict(allowed_props) else allowed_props
    result = allowed.get(key)
    if result is None:
        raise KeyError(f"not found allowed key from {allowed}")
    return result



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from typing import Callable, List


class AbstractCallbackManager(metaclass=ABCMeta):

    def __init__(self):
        self._callbacks: List[Callable] = []

    @property
    def callbacks(self):
        return self._callbacks

    @property
    def length(self):
        return len(self.callbacks)

    def empty(self):
        return not self.length

    def add(self, function: Callable) -> int:
        self.check(function)
        if not self._is_exits(function):
            self.callbacks.append(function)
        return self.length

    def remove(self, function: Callable) -> int:
        if self.empty():
            return self.length
        idx = self._find(function)
        if idx > -1:
            self.callbacks.pop(idx)
        return self.length

    def __add__(self, other):
        if self.is_callback_manager(other):
            self.callbacks.extend(other.callbacks)
        return self

    def __iter__(self):
        if not self.empty():
            for _, cb in enumerate(self.callbacks):
                yield cb

    @abstractmethod
    def check(self, value) -> bool:
        raise NotImplementedError("NotImplemented .check(value) -> bool")

    def is_callback_manager(self, value):
        return isinstance(value, AbstractCallbackManager)

    def _find(self, function):
        if not self._is_exits(function):
            return -1
        return self.callbacks.index(function)

    def _is_exits(self, function):
        if self.empty():
            return False
        return function in self.callbacks



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
try:
    import simplejson as json
except ImportError:
    import json
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from configparser import ConfigParser
from enum import Enum
from typing import Dict, DefaultDict

from ..utils import (
    ExtensionNameList,
    EnumUtil,
    Encoding,
    get_from_allowed,
    PathUtil,
    PathTypes,
    StringsUtil
)
from ..validator.is_validators import (
    is_any_string,
    is_dict,
    is_string,
    is_int,
    is_float,
    is_number
)


class AllowedOptions(Enum):
    pass


class AbstractConfigReader(metaclass=ABCMeta):
    """
    config reader abc class
    """

    def __init__(self,
                 filename: PathTypes, *,
                 options: AllowedOptions,
                 extension_names: ExtensionNameList = None):
        self._file = filename
        self._options = self._to_options(options)
        self._extension_names = [] if not extension_names else extension_names

    @property
    def file(self):
        return self._file

    @property
    def extension_names(self) -> ExtensionNameList:
        return self._extension_names

    @abstractmethod
    def get_int(self, node: str, option) -> int:
        raise NotImplementedError("NotImplemented .get_int(node, option) -> int")

    @abstractmethod
    def get_float(self, node: str, option) -> float:
        raise NotImplementedError("NotImplemented .get_float(node, option) -> float")

    @abstractmethod
    def get_string(self, node: str, option) -> str:
        raise NotImplementedError("NotImplemented .get_string(node, option) -> str")

    def _to_int(self, value):
        if is_int(value):
            return value
        elif is_number(value):
            return int(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    def _to_float(self, value):
        if is_float(value):
            return value
        elif is_number(value):
            return float(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    def _to_string(self, value):
        if is_string(value):
            return value
        elif is_any_string(value):
            return StringsUtil.anystr_to_string(value)
        return str(value)

    def _get_allowed_option(self, option: AllowedOptions) -> str:
        result = get_from_allowed(option, self._options)
        if not is_string(result):
            raise ValueError(f"option need string, but got {type(result).__name__}")
        return result

    def _to_options(self, options: AllowedOptions) -> dict:
        return EnumUtil.create_dict_by_enum(options)

    def is_config_file(self, file: PathTypes) -> bool:
        return PathUtil.is_endswith_ext(file, self.extension_names)


class IniReader(AbstractConfigReader):
    DEFAULT = "DEFAULT"

    def __init__(self,
                 filename: PathTypes,
                 *,
                 options: AllowedOptions,
                 extension_names: ExtensionNameList = None):
        super().__init__(
            filename,
            options=options,
            extension_names=["ini"]
        )
        self._parser = self._init_parser(self.file)

    def get_float(self, node: str, option: AllowedOptions) -> float:
        return self._parser.getfloat(
            section=self._get_node(node),
            option=self._get_allowed_option(option)
        )

    def get_int(self, node: str, option: AllowedOptions) -> int:
        return self._parser.getint(
            section=self._get_node(node),
            option=self._get_allowed_option(option)
        )

    def get_string(self, node: str, option: AllowedOptions) -> str:
        result = self._parser.get(
            section=self._get_node(node),
            option=self._get_allowed_option(option)
        )
        return self._strip_quota(result)

    def has_node(self, node):
        if node is None:
            return False
        return self._parser.has_section(node)

    def _strip_quota(self, result: str) -> str:
        if not is_string(result):
            raise ValueError(f"result need string, but got {type(result).__name__}")
        if not result:
            return result
        return result.strip("'").strip('"')

    def _get_node(self, node: str):
        return IniReader.DEFAULT if not self.has_node(node) else node

    def _init_parser(self, file: PathTypes) -> ConfigParser:
        parser = ConfigParser()
        self._read_ini(parser, file)
        return parser

    def _read_ini(self, parser: ConfigParser, file: PathTypes):
        if self.is_config_file(file):
            parser.read(file)


class JsonReader(AbstractConfigReader):
    class JsonNode(object):

        def __init__(self, **kwargs):
            self._node = {} if not kwargs else kwargs

        def get(self, item, default=None):
            return self._node.get(item, default)

    def __init__(self,
                 filename: PathTypes,
                 *,
                 options: AllowedOptions,
                 extension_names: ExtensionNameList = None,
                 encoding: Encoding = Encoding.UTF_8):
        super().__init__(
            filename, options=options, extension_names=["json"]
        )
        self._parser = self._init_parser(self.file, encoding)

    def get_float(self, node: str, option: AllowedOptions) -> float:
        return self._get_from_parser(node, option, self._to_float)

    def get_int(self, node: str, option: AllowedOptions) -> int:
        return self._get_from_parser(node, option, self._to_int)

    def get_string(self, node: str, option: AllowedOptions) -> str:
        return self._get_from_parser(node, option, self._to_string)

    def _get_from_parser(self, node, option: AllowedOptions, callback):
        json_node = self._get_node(node)
        allowed = self._get_allowed_option(option)
        result = json_node.get(allowed)
        if result is None:
            raise KeyError(f"not found {allowed} from {node}")
        return callback(result)

    def _get_node(self, node):
        return self._parser[node]

    def _init_parser(self,
                     file: PathTypes,
                     encoding: Encoding = Encoding.UTF_8) -> DefaultDict[str, JsonNode]:
        data = self._read_json(file, encoding)
        filtered = self._filter_json(data)
        return defaultdict(JsonReader.JsonNode, filtered)

    def _filter_json(self, result: dict) -> Dict[str, JsonNode]:
        if not is_dict(result):
            return {}
        return {k: JsonReader(**v) for k, v in result.items() if is_dict(v)}

    def _read_json(self, file, encoding: Encoding) -> dict:
        result = {}
        if self.is_config_file(file):
            with open(file, mode="r") as jf:
                result = json.load(jf, encoding=StringsUtil.to_encoding(encoding))
        return result



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum


def is_enum(value):
    return isinstance(value, Enum)


def is_enum_subclass(klass):
    return issubclass(klass, Enum)



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .src import validator
from .src import utils
from .src import callback
from .src import conf



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from functools import partial

from .reader import AllowedOptions, AbstractConfigReader
from ..utils import ExtensionNameList, PathTypes
from ..validator.is_validators import is_iterable, is_list


class IFactory(object):

    @classmethod
    def create(cls, *args, **kwargs) -> AbstractConfigReader:
        raise NotImplementedError("")


class ReaderFactory(IFactory):

    @classmethod
    def create(cls,
               subclass: AbstractConfigReader,
               file: PathTypes,
               options: AllowedOptions,
               extension_names: ExtensionNameList = None,
               **kwargs) -> AbstractConfigReader:
        f = cls._factory(
            subclass, file, options, extension_names
        )
        return f(**kwargs)

    @classmethod
    def _factory(cls,
                 subclass: AbstractConfigReader,
                 file: PathTypes,
                 options: AllowedOptions,
                 extension_names: ExtensionNameList = None):
        if not cls.is_reader_subclass(subclass):
            raise ValueError(f"need a AbstractConfigReader subclass, but got {type(subclass).__name__}")
        klass = subclass
        return partial(
            klass,
            file,
            options=options,
            extension_names=cls.to_extension_names(extension_names)
        )

    @staticmethod
    def to_extension_names(extension_names: ExtensionNameList = None) -> list:
        if is_list(extension_names):
            return extension_names
        elif is_iterable(extension_names):
            return list(extension_names)
        else:
            return []

    @staticmethod
    def is_reader_subclass(subclass):
        return issubclass(subclass, AbstractConfigReader)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from ._is_dict import is_dict
from ._is_enum import is_enum, is_enum_subclass
from ._is_function import is_coroutine, is_function, is_async_function
from ._is_iterable import is_iterable
from ._is_list import is_list
from ._is_num import is_int, is_float, is_number
from ._is_path import is_path
from ._is_strings import is_bytes, is_bytearray, is_string, is_any_string
